@@ -18,6 +18,10 @@
 #include "apache2.h"
 #include "msc_crypt.h"
 
+#ifdef APLOG_USE_MODULE
+    APLOG_USE_MODULE(security2);
+#endif
+
 /* -- Input filter -- */
 
 #if 0
@@ -36,6 +40,7 @@ apr_status_t input_filter(ap_filter_t *f, apr_bucket_brigade *bb_out,
     msc_data_chunk *chunk = NULL;
     apr_bucket *bucket;
     apr_status_t rc;
+    int no_data = 1;
     char *my_error_msg = NULL;
 
     if (msr == NULL) {
@@ -85,10 +90,11 @@ apr_status_t input_filter(ap_filter_t *f, apr_bucket_brigade *bb_out,
         return APR_EGENERAL;
     }
 
-    if (chunk && (!msr->txcfg->stream_inbody_inspection || (msr->txcfg->stream_inbody_inspection && msr->if_stream_changed == 0))) {
-        /* Copy the data we received in the chunk */
-        bucket = apr_bucket_heap_create(chunk->data, chunk->length, NULL,
-                f->r->connection->bucket_alloc);
+    if (chunk && chunk->length > 0) {
+        if (chunk && (!msr->txcfg->stream_inbody_inspection || (msr->txcfg->stream_inbody_inspection && msr->if_stream_changed == 0))) {
+            /* Copy the data we received in the chunk */
+            bucket = apr_bucket_heap_create(chunk->data, chunk->length, NULL,
+                    f->r->connection->bucket_alloc);
 
 #if 0
 
@@ -107,44 +113,50 @@ apr_status_t input_filter(ap_filter_t *f, apr_bucket_brigade *bb_out,
 
 #endif
 
-        if (bucket == NULL) return APR_EGENERAL;
-        APR_BRIGADE_INSERT_TAIL(bb_out, bucket);
+            if (bucket == NULL) return APR_EGENERAL;
+            APR_BRIGADE_INSERT_TAIL(bb_out, bucket);
+            no_data = 0;
 
-        if (msr->txcfg->debuglog_level >= 4) {
-            msr_log(msr, 4, "Input filter: Forwarded %" APR_SIZE_T_FMT " bytes.", chunk->length);
-        }
-    } else if (msr->stream_input_data != NULL) {
-
-        msr->if_stream_changed = 0;
-
-        bucket = apr_bucket_heap_create(msr->stream_input_data, msr->stream_input_length, NULL,
-                f->r->connection->bucket_alloc);
-
-        if (msr->txcfg->stream_inbody_inspection)  {
-            if(msr->stream_input_data != NULL) {
-                free(msr->stream_input_data);
-                msr->stream_input_data = NULL;
+            if (msr->txcfg->debuglog_level >= 4) {
+                msr_log(msr, 4, "Input filter: Forwarded %" APR_SIZE_T_FMT " bytes.", chunk->length);
             }
+        } else if (msr->stream_input_data != NULL) {
+
+            msr->if_stream_changed = 0;
+
+            bucket = apr_bucket_heap_create(msr->stream_input_data, msr->stream_input_length, NULL,
+                    f->r->connection->bucket_alloc);
+
+            if (msr->txcfg->stream_inbody_inspection)  {
+                if(msr->stream_input_data != NULL) {
+                    free(msr->stream_input_data);
+                    msr->stream_input_data = NULL;
+                }
+            }
+
+            if (bucket == NULL) return APR_EGENERAL;
+            APR_BRIGADE_INSERT_TAIL(bb_out, bucket);
+            no_data = 0;
+
+            if (msr->txcfg->debuglog_level >= 4) {
+                msr_log(msr, 4, "Input stream filter: Forwarded %" APR_SIZE_T_FMT " bytes.", msr->stream_input_length);
+            }
+
         }
-
-        if (bucket == NULL) return APR_EGENERAL;
-        APR_BRIGADE_INSERT_TAIL(bb_out, bucket);
-
-        if (msr->txcfg->debuglog_level >= 4) {
-            msr_log(msr, 4, "Input stream filter: Forwarded %" APR_SIZE_T_FMT " bytes.", msr->stream_input_length);
-        }
-
     }
 
     if (rc == 0) {
         modsecurity_request_body_retrieve_end(msr);
 
-        bucket = apr_bucket_eos_create(f->r->connection->bucket_alloc);
-        if (bucket == NULL) return APR_EGENERAL;
-        APR_BRIGADE_INSERT_TAIL(bb_out, bucket);
+        if (msr->if_seen_eos) {
+            bucket = apr_bucket_eos_create(f->r->connection->bucket_alloc);
+            if (bucket == NULL) return APR_EGENERAL;
+            APR_BRIGADE_INSERT_TAIL(bb_out, bucket);
+            no_data = 0;
 
-        if (msr->txcfg->debuglog_level >= 4) {
-            msr_log(msr, 4, "Input filter: Sent EOS.");
+            if (msr->txcfg->debuglog_level >= 4) {
+                msr_log(msr, 4, "Input filter: Sent EOS.");
+            }
         }
 
         /* We're done */
@@ -154,6 +166,10 @@ apr_status_t input_filter(ap_filter_t *f, apr_bucket_brigade *bb_out,
         if (msr->txcfg->debuglog_level >= 4) {
             msr_log(msr, 4, "Input filter: Input forwarding complete.");
         }
+
+        if (no_data) {
+            return ap_get_brigade(f->next, bb_out, mode, block, nbytes);
+        }
     }
 
     return APR_SUCCESS;
@@ -164,7 +180,7 @@ apr_status_t input_filter(ap_filter_t *f, apr_bucket_brigade *bb_out,
  */
 apr_status_t read_request_body(modsec_rec *msr, char **error_msg) {
     request_rec *r = msr->r;
-    unsigned int seen_eos;
+    unsigned int finished_reading;
     apr_bucket_brigade *bb_in;
     apr_bucket *bucket;
 
@@ -188,12 +204,12 @@ apr_status_t read_request_body(modsec_rec *msr, char **error_msg) {
     if (msr->txcfg->debuglog_level >= 4) {
         msr_log(msr, 4, "Input filter: Reading request body.");
     }
-
     if (modsecurity_request_body_start(msr, error_msg) < 0) {
         return -1;
     }
 
-    seen_eos = 0;
+    finished_reading = 0;
+    msr->if_seen_eos = 0;
     bb_in = apr_brigade_create(msr->mp, r->connection->bucket_alloc);
     if (bb_in == NULL) return -1;
     do {
@@ -205,6 +221,9 @@ apr_status_t read_request_body(modsec_rec *msr, char **error_msg) {
              *      too large and APR_EGENERAL when the client disconnects.
              */
             switch(rc) {
+                case APR_INCOMPLETE :
+                    *error_msg = apr_psprintf(msr->mp, "Error reading request body: %s", get_apr_error(msr->mp, rc));
+                    return -7;
                 case APR_EOF :
                     *error_msg = apr_psprintf(msr->mp, "Error reading request body: %s", get_apr_error(msr->mp, rc));
                     return -6;
@@ -275,14 +294,25 @@ apr_status_t read_request_body(modsec_rec *msr, char **error_msg) {
             }
 
             if (msr->txcfg->stream_inbody_inspection == 1)   {
+#ifndef MSC_LARGE_STREAM_INPUT
                 msr->stream_input_length+=buflen;
                 modsecurity_request_body_to_stream(msr, buf, buflen, error_msg);
+#else
+                if (modsecurity_request_body_to_stream(msr, buf, buflen, error_msg) < 0) {
+                    return -1;
+                }
+#endif
             }
 
             msr->reqbody_length += buflen;
 
             if (buflen != 0) {
                 int rcbs = modsecurity_request_body_store(msr, buf, buflen, error_msg);
+
+                if (msr->reqbody_length > (apr_size_t)msr->txcfg->reqbody_limit && msr->txcfg->if_limit_action == REQUEST_BODY_LIMIT_ACTION_PARTIAL) {
+                    finished_reading = 1;
+                }
+
                 if (rcbs < 0) {
                     if (rcbs == -5) {
                         if((msr->txcfg->is_enabled == MODSEC_ENABLED) && (msr->txcfg->if_limit_action == REQUEST_BODY_LIMIT_ACTION_REJECT)) {
@@ -309,15 +339,15 @@ apr_status_t read_request_body(modsec_rec *msr, char **error_msg) {
             }
 
             if (APR_BUCKET_IS_EOS(bucket)) {
-                seen_eos = 1;
+                finished_reading = 1;
+                msr->if_seen_eos = 1;
             }
         }
 
         apr_brigade_cleanup(bb_in);
-    } while(!seen_eos);
+    } while(!finished_reading);
 
-    // TODO: Why ignore the return code here?
-    modsecurity_request_body_end(msr, error_msg);
+    apr_status_t rcbe = modsecurity_request_body_end(msr, error_msg);
 
     if (msr->txcfg->debuglog_level >= 4) {
         msr_log(msr, 4, "Input filter: Completed receiving request body (length %" APR_SIZE_T_FMT ").",
@@ -326,7 +356,7 @@ apr_status_t read_request_body(modsec_rec *msr, char **error_msg) {
 
     msr->if_status = IF_STATUS_WANTS_TO_RUN;
 
-    return 1;
+    return rcbe;
 }
 
 
@@ -600,8 +630,12 @@ static int flatten_response_body(modsec_rec *msr) {
             retval = hash_response_body_links(msr);
             if(retval > 0) {
                 retval = inject_hashed_response_body(msr, retval);
-                if (msr->txcfg->debuglog_level >= 4) {
-                    msr_log(msr, 4, "Hash completed in %" APR_TIME_T_FMT " usec.", (apr_time_now() - time1));
+                if(retval < 0){
+                    msr_log(msr, 1, "inject_hashed_response_body: Unable to inject hash into response body. Returning response without changes." );
+                }else{
+                    if (msr->txcfg->debuglog_level >= 4) {
+                        msr_log(msr, 4, "Hash completed in %" APR_TIME_T_FMT " usec.", (apr_time_now() - time1));
+                    }
                 }
 
             }
