@@ -1,6 +1,6 @@
 /*
 * ModSecurity for Apache 2.x, http://www.modsecurity.org/
-* Copyright (c) 2004-2013 Trustwave Holdings, Inc. (http://www.trustwave.com/)
+* Copyright (c) 2004-2022 Trustwave Holdings, Inc. (http://www.trustwave.com/)
 *
 * You may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
@@ -19,9 +19,18 @@
 #include "modsecurity.h"
 #include "msc_parsers.h"
 #include "msc_util.h"
+#include "msc_json.h"
 #include "msc_xml.h"
 #include "apr_version.h"
 
+#ifdef WITH_CURL
+#include <curl/curl.h>
+#endif
+
+#ifdef APLOG_USE_MODULE
+    APLOG_USE_MODULE(security2);
+#endif
+
 unsigned long int DSOLOCAL unicode_codepage = 0;
 
 int DSOLOCAL *unicode_map_table = NULL;
@@ -117,8 +126,19 @@ msc_engine *modsecurity_create(apr_pool_t *mp, int processing_mode) {
 int modsecurity_init(msc_engine *msce, apr_pool_t *mp) {
     apr_status_t rc;
 
+    /**
+     * Notice that curl is initialized here but never cleaned up. First version
+     * of this implementation curl was initialized and cleaned for every
+     * utilization. Turns out that it was not only cleaning stuff that was
+     * utilized by Curl but also other OpenSSL stuff that was utilized by
+     * mod_ssl leading the SSL support to crash.
+     */
+#ifdef WITH_CURL
+    curl_global_init(CURL_GLOBAL_ALL);
+#endif
     /* Serial audit log mutext */
-    rc = apr_global_mutex_create(&msce->auditlog_lock, NULL, APR_LOCK_DEFAULT, mp);
+    tmpnam(auditlog_lock_name);
+    rc = apr_global_mutex_create(&msce->auditlog_lock, auditlog_lock_name, APR_LOCK_DEFAULT, mp);
     if (rc != APR_SUCCESS) {
         //ap_log_error(APLOG_MARK, APLOG_ERR, rv, s, "mod_security: Could not create modsec_auditlog_lock");
         //return HTTP_INTERNAL_SERVER_ERROR;
@@ -139,7 +159,8 @@ int modsecurity_init(msc_engine *msce, apr_pool_t *mp) {
     }
 #endif /* SET_MUTEX_PERMS */
 
-    rc = apr_global_mutex_create(&msce->geo_lock, NULL, APR_LOCK_DEFAULT, mp);
+    tmpnam(geo_lock_name);
+    rc = apr_global_mutex_create(&msce->geo_lock, geo_lock_name, APR_LOCK_DEFAULT, mp);
     if (rc != APR_SUCCESS) {
         return -1;
     }
@@ -154,6 +175,25 @@ int modsecurity_init(msc_engine *msce, apr_pool_t *mp) {
         return -1;
     }
 #endif /* SET_MUTEX_PERMS */
+
+#ifdef GLOBAL_COLLECTION_LOCK
+    tmpnam(dbm_lock_name);
+    rc = apr_global_mutex_create(&msce->dbm_lock, dbm_lock_name, APR_LOCK_DEFAULT, mp);
+    if (rc != APR_SUCCESS) {
+        return -1;
+    }
+
+#ifdef __SET_MUTEX_PERMS
+#if AP_SERVER_MAJORVERSION_NUMBER > 1 && AP_SERVER_MINORVERSION_NUMBER > 2
+    rc = ap_unixd_set_global_mutex_perms(msce->dbm_lock);
+#else
+    rc = unixd_set_global_mutex_perms(msce->dbm_lock);
+#endif
+    if (rc != APR_SUCCESS) {
+        return -1;
+    }
+#endif /* SET_MUTEX_PERMS */
+#endif
 #endif
 
     return 1;
@@ -180,6 +220,15 @@ void modsecurity_child_init(msc_engine *msce) {
         }
     }
 
+#ifdef GLOBAL_COLLECTION_LOCK
+    if (msce->dbm_lock != NULL) {
+        apr_status_t rc = apr_global_mutex_child_init(&msce->dbm_lock, NULL, msce->mp);
+        if (rc != APR_SUCCESS) {
+            // ap_log_error(APLOG_MARK, APLOG_ERR, rs, s, "Failed to child-init dbm mutex");
+        }
+    }
+#endif
+
 }
 
 /**
@@ -222,9 +271,11 @@ static void modsecurity_persist_data(modsec_rec *msr) {
     }
 
     /* Remove stale collections. */
-    srand(time(NULL));
-
+#if AP_SERVER_MAJORVERSION_NUMBER > 1 && AP_SERVER_MINORVERSION_NUMBER > 3
+    if (ap_random_pick(0, RAND_MAX) < RAND_MAX/100) {
+#else
     if (rand() < RAND_MAX/100) {
+#endif
         arr = apr_table_elts(msr->collections);
         te = (apr_table_entry_t *)arr->elts;
         for (i = 0; i < arr->nelts; i++) {
@@ -255,18 +306,44 @@ static apr_status_t modsecurity_tx_cleanup(void *data) {
     /* XML processor cleanup. */
     if (msr->xml != NULL) xml_cleanup(msr);
 
+#ifdef WITH_YAJL
+    /* JSON processor cleanup. */
+    if (msr->json != NULL) json_cleanup(msr);
+#endif
+
     // TODO: Why do we ignore return code here?
     modsecurity_request_body_clear(msr, &my_error_msg);
     if (my_error_msg != NULL) {
         msr_log(msr, 1, "%s", my_error_msg);
     }
 
+    if (msr->msc_full_request_length > 0 &&
+            msr->msc_full_request_buffer != NULL) {
+        msr->msc_full_request_length = 0;
+        free(msr->msc_full_request_buffer);
+    }
+
 #if defined(WITH_LUA)
     #ifdef CACHE_LUA
     if(msr->L != NULL)  lua_close(msr->L);
     #endif
 #endif
 
+    /* Streams cleanup. */
+    if (msr->stream_input_data != NULL) {
+        free(msr->stream_input_data);
+        msr->stream_input_data = NULL;
+        msr->stream_input_length = 0;
+#ifdef MSC_LARGE_STREAM_INPUT
+        msr->stream_input_allocated_length = 0;
+#endif
+    }
+    if (msr->stream_output_data != NULL) {
+        free(msr->stream_output_data);
+        msr->stream_output_data = NULL;
+        msr->stream_output_length = 0;
+    }
+
     return APR_SUCCESS;
 }
 
@@ -488,7 +565,11 @@ static int is_response_status_relevant(modsec_rec *msr, int status) {
 
     rc = msc_regexec(msr->txcfg->auditlog_relevant_regex, buf, strlen(buf), &my_error_msg);
     if (rc >= 0) return 1;
+#ifdef WITH_PCRE2
+    if (rc == PCRE2_ERROR_NOMATCH) return 0;
+#else
     if (rc == PCRE_ERROR_NOMATCH) return 0;
+#endif
 
     msr_log(msr, 1, "Regex processing failed (rc %d): %s", rc, my_error_msg);
     
@@ -524,7 +605,7 @@ static apr_status_t modsecurity_process_phase_request_body(modsec_rec *msr) {
     apr_time_t time_before;
     apr_status_t rc = 0;
 
-    
+
     if ((msr->allow_scope == ACTION_ALLOW_REQUEST)||(msr->allow_scope == ACTION_ALLOW)) {
         if (msr->txcfg->debuglog_level >= 4) {
             msr_log(msr, 4, "Skipping phase REQUEST_BODY (allow used).");
@@ -614,7 +695,7 @@ static apr_status_t modsecurity_process_phase_response_body(modsec_rec *msr) {
  */
 static apr_status_t modsecurity_process_phase_logging(modsec_rec *msr) {
     apr_time_t time_before, time_after;
-    
+
     if (msr->txcfg->debuglog_level >= 4) {
         msr_log(msr, 4, "Starting phase LOGGING.");
     }
