@@ -1,6 +1,6 @@
 /*
  * ModSecurity for Apache 2.x, http://www.modsecurity.org/
- * Copyright (c) 2004-2013 Trustwave Holdings, Inc. (http://www.trustwave.com/)
+ * Copyright (c) 2004-2022 Trustwave Holdings, Inc. (http://www.trustwave.com/)
  *
  * You may not use this file except in compliance with
  * the License.  You may obtain a copy of the License at
@@ -67,25 +67,31 @@ char *normalize_path(modsec_rec *msr, char *input) {
         if(uri->path)   {
             char *Uri = NULL;
             int bytes = 0;
-            int i;
-            char *relative_link = NULL;
+            /*int i;*/
+            char *abs_link = NULL;
             char *filename = NULL;
-            char *relative_path = NULL;
-            char *relative_uri = NULL;
+            char *abs_path = NULL;
+            char *abs_uri = NULL;
 
-            filename = file_basename(msr->mp, msr->r->parsed_uri.path);
+            if (uri->path[0] != '/') {
+                /* uri->path is relative. make it absolute */
+                filename = file_basename(msr->mp, msr->r->parsed_uri.path);
 
-            if(filename == NULL || (strlen(msr->r->parsed_uri.path) - strlen(filename) < 0))
-                return NULL;
+                if(filename == NULL || (strlen(msr->r->parsed_uri.path) - strlen(filename) < 0))
+                    return NULL;
 
-            relative_path = apr_pstrndup(msr->mp, msr->r->parsed_uri.path, strlen(msr->r->parsed_uri.path) - strlen(filename));
-            relative_uri = apr_pstrcat(msr->mp, relative_path, uri->path, NULL);
+                abs_path = apr_pstrndup(msr->mp, msr->r->parsed_uri.path, strlen(msr->r->parsed_uri.path) - strlen(filename));
+                abs_uri = apr_pstrcat(msr->mp, abs_path, uri->path, NULL);
 
-            relative_link = apr_pstrdup(msr->mp, relative_uri);
+                abs_link = apr_pstrdup(msr->mp, abs_uri);
+            }
+            else {
+                abs_link = apr_pstrdup(msr->mp, uri->path);
+            }
 
-            xmlNormalizeURIPath(relative_link);
+            xmlNormalizeURIPath(abs_link);
 
-            Uri = apr_pstrdup(msr->mp, relative_link);
+            Uri = apr_pstrdup(msr->mp, abs_link);
 
 /*
             for(i = 0; i < (int)strlen(Uri); i++)    {
@@ -152,14 +158,14 @@ char *getkey(apr_pool_t *mp) {
     char salt[64];
 
     apr_generate_random_bytes(salt, sizeof(salt));
-    key = apr_psprintf(mp,"%s",salt);
+    key = apr_psprintf(mp,"%.*s",(int)sizeof(salt),salt);
 
     apr_sha1_init (&ctx);
     apr_sha1_update (&ctx, (const char*)key, strlen(key));
     apr_sha1_update (&ctx, "\0", 1);
 
     apr_generate_random_bytes(salt, sizeof(salt));
-    value = apr_psprintf(mp,"%s",salt);
+    value = apr_psprintf(mp,"%.*s",(int)sizeof(salt),salt);
 
     apr_sha1_update (&ctx, value, strlen (value));
     apr_sha1_final (digest, &ctx);
@@ -182,14 +188,19 @@ char *getkey(apr_pool_t *mp) {
  *
  * \retval hex_digest The MAC
  */
+#ifdef __NetBSD__
+char *mschmac(modsec_rec *msr, const char *key, int key_len,
+        unsigned char *msg, int msglen) {
+#else
 char *hmac(modsec_rec *msr, const char *key, int key_len,
         unsigned char *msg, int msglen) {
+#endif
     apr_sha1_ctx_t ctx;
     unsigned char digest[APR_SHA1_DIGESTSIZE];
     unsigned char hmac_ipad[HMAC_PAD_SIZE], hmac_opad[HMAC_PAD_SIZE];
     unsigned char nkey[APR_SHA1_DIGESTSIZE];
     unsigned char *hmac_key = (unsigned char *) key;
-    char hex_digest[APR_SHA1_DIGESTSIZE * 2], *hmac_digest;
+    char hex_digest[APR_SHA1_DIGESTSIZE * 2 + 1], *hmac_digest;
     const char hex[] = "0123456789abcdef";
     int i;
 
@@ -375,7 +386,11 @@ int do_hash_method(modsec_rec *msr, char *link, int type)   {
                 case HASH_URL_HREF_HASH_RX:
                     if(em[i]->type == HASH_URL_HREF_HASH_RX)   {
                         rc = msc_regexec_capture(em[i]->param_data, link, strlen(link), ovector, 30, &my_error_msg);
+#ifdef WITH_PCRE2
+                        if ((rc == PCRE2_ERROR_MATCHLIMIT) || (rc == PCRE2_ERROR_RECURSIONLIMIT)) {
+#else
                         if ((rc == PCRE_ERROR_MATCHLIMIT) || (rc == PCRE_ERROR_RECURSIONLIMIT)) {
+#endif
                             msc_string *s = (msc_string *)apr_pcalloc(msr->mp, sizeof(msc_string));
 
                             if (s == NULL) return -1;
@@ -404,7 +419,11 @@ int do_hash_method(modsec_rec *msr, char *link, int type)   {
                                 msr_log(msr, 4, "%s.", error_msg);
                             return -1;
                         }
+#ifdef WITH_PCRE2
+                        if (rc != PCRE2_ERROR_NOMATCH) { /* Match. */
+#else
                         if (rc != PCRE_ERROR_NOMATCH) { /* Match. */
+#endif
                             return 1;
                         }
                     }
@@ -430,7 +449,11 @@ int do_hash_method(modsec_rec *msr, char *link, int type)   {
                 case HASH_URL_FACTION_HASH_RX:
                     if(em[i]->type == HASH_URL_FACTION_HASH_RX)   {
                         rc = msc_regexec_capture(em[i]->param_data, link, strlen(link), ovector, 30, &my_error_msg);
+#ifdef WITH_PCRE2
+                        if ((rc == PCRE2_ERROR_MATCHLIMIT) || (rc == PCRE2_ERROR_RECURSIONLIMIT)) {
+#else
                         if ((rc == PCRE_ERROR_MATCHLIMIT) || (rc == PCRE_ERROR_RECURSIONLIMIT)) {
+#endif
                             msc_string *s = (msc_string *)apr_pcalloc(msr->mp, sizeof(msc_string));
 
                             if (s == NULL) return -1;
@@ -459,7 +482,11 @@ int do_hash_method(modsec_rec *msr, char *link, int type)   {
                                 msr_log(msr, 4, "%s.", error_msg);
                             return -1;
                         }
+#ifdef WITH_PCRE2
+                        if (rc != PCRE2_ERROR_NOMATCH) { /* Match. */
+#else
                         if (rc != PCRE_ERROR_NOMATCH) { /* Match. */
+#endif
                             return 1;
                         }
                     }
@@ -485,7 +512,11 @@ int do_hash_method(modsec_rec *msr, char *link, int type)   {
                 case HASH_URL_LOCATION_HASH_RX:
                     if(em[i]->type == HASH_URL_LOCATION_HASH_RX)   {
                         rc = msc_regexec_capture(em[i]->param_data, link, strlen(link), ovector, 30, &my_error_msg);
+#ifdef WITH_PCRE2
+                        if ((rc == PCRE2_ERROR_MATCHLIMIT) || (rc == PCRE2_ERROR_RECURSIONLIMIT)) {
+#else
                         if ((rc == PCRE_ERROR_MATCHLIMIT) || (rc == PCRE_ERROR_RECURSIONLIMIT)) {
+#endif
                             msc_string *s = (msc_string *)apr_pcalloc(msr->mp, sizeof(msc_string));
 
                             if (s == NULL) return -1;
@@ -514,7 +545,11 @@ int do_hash_method(modsec_rec *msr, char *link, int type)   {
                                 msr_log(msr, 4, "%s.", error_msg);
                             return -1;
                         }
+#ifdef WITH_PCRE2
+                        if (rc != PCRE2_ERROR_NOMATCH) { /* Match. */
+#else
                         if (rc != PCRE_ERROR_NOMATCH) { /* Match. */
+#endif
                             return 1;
                         }
                     }
@@ -540,7 +575,11 @@ int do_hash_method(modsec_rec *msr, char *link, int type)   {
                 case HASH_URL_IFRAMESRC_HASH_RX:
                     if(em[i]->type == HASH_URL_IFRAMESRC_HASH_RX)   {
                         rc = msc_regexec_capture(em[i]->param_data, link, strlen(link), ovector, 30, &my_error_msg);
+#ifdef WITH_PCRE2
+                        if ((rc == PCRE2_ERROR_MATCHLIMIT) || (rc == PCRE2_ERROR_RECURSIONLIMIT)) {
+#else
                         if ((rc == PCRE_ERROR_MATCHLIMIT) || (rc == PCRE_ERROR_RECURSIONLIMIT)) {
+#endif
                             msc_string *s = (msc_string *)apr_pcalloc(msr->mp, sizeof(msc_string));
 
                             if (s == NULL) return -1;
@@ -569,7 +608,11 @@ int do_hash_method(modsec_rec *msr, char *link, int type)   {
                                 msr_log(msr, 4, "%s.", error_msg);
                             return -1;
                         }
+#ifdef WITH_PCRE2
+                        if (rc != PCRE2_ERROR_NOMATCH) { /* Match. */
+#else
                         if (rc != PCRE_ERROR_NOMATCH) { /* Match. */
+#endif
                             return 1;
                         }
                     }
@@ -595,7 +638,11 @@ int do_hash_method(modsec_rec *msr, char *link, int type)   {
                 case HASH_URL_FRAMESRC_HASH_RX:
                     if(em[i]->type == HASH_URL_FRAMESRC_HASH_RX)   {
                         rc = msc_regexec_capture(em[i]->param_data, link, strlen(link), ovector, 30, &my_error_msg);
+#ifdef WITH_PCRE2
+                        if ((rc == PCRE2_ERROR_MATCHLIMIT) || (rc == PCRE2_ERROR_RECURSIONLIMIT)) {
+#else
                         if ((rc == PCRE_ERROR_MATCHLIMIT) || (rc == PCRE_ERROR_RECURSIONLIMIT)) {
+#endif
                             msc_string *s = (msc_string *)apr_pcalloc(msr->mp, sizeof(msc_string));
 
                             if (s == NULL) return -1;
@@ -624,7 +671,11 @@ int do_hash_method(modsec_rec *msr, char *link, int type)   {
                                 msr_log(msr, 4, "%s.", error_msg);
                             return -1;
                         }
+#ifdef WITH_PCRE2
+                        if (rc != PCRE2_ERROR_NOMATCH) { /* Match. */
+#else
                         if (rc != PCRE_ERROR_NOMATCH) { /* Match. */
+#endif
                             return 1;
                         }
                     }
@@ -1072,7 +1123,8 @@ int inject_hashed_response_body(modsec_rec *msr, int elts) {
     }
 
     htmlDocContentDumpFormatOutput(output_buf, msr->crypto_html_tree, NULL, 0);
-    xmlOutputBufferFlush(output_buf);
+    // Not necessary in 2.9.4+
+    //xmlOutputBufferFlush(output_buf);
 
 #ifdef  LIBXML2_NEW_BUFFER
 
@@ -1082,6 +1134,7 @@ int inject_hashed_response_body(modsec_rec *msr, int elts) {
             xmlOutputBufferClose(output_buf);
             xmlFreeDoc(msr->crypto_html_tree);
             msr->of_stream_changed = 0;
+            msr_log(msr, 4, "inject_hashed_response_body: NEW_BUFFER Output buffer is null.");
             return -1;
         }
 
@@ -1096,6 +1149,7 @@ int inject_hashed_response_body(modsec_rec *msr, int elts) {
         if (msr->stream_output_data == NULL) {
             xmlOutputBufferClose(output_buf);
             xmlFreeDoc(msr->crypto_html_tree);
+            msr_log(msr, 4, "inject_hashed_response_body: NEW BUFFER Stream Output is null.");
             return -1;
         }
 
@@ -1103,7 +1157,7 @@ int inject_hashed_response_body(modsec_rec *msr, int elts) {
         memcpy(msr->stream_output_data, xmlOutputBufferGetContent(output_buf), msr->stream_output_length);
 
         if (msr->txcfg->debuglog_level >= 4)
-            msr_log(msr, 4, "inject_hashed_response_body: Copying XML tree from CONTENT to stream buffer [%d] bytes.", xmlOutputBufferGetSize(output_buf));
+            msr_log(msr, 4, "inject_hashed_response_body: Copying XML tree from CONTENT to stream buffer [%zu] bytes.", xmlOutputBufferGetSize(output_buf));
 
     } else {
 
@@ -1111,6 +1165,7 @@ int inject_hashed_response_body(modsec_rec *msr, int elts) {
             xmlOutputBufferClose(output_buf);
             xmlFreeDoc(msr->crypto_html_tree);
             msr->of_stream_changed = 0;
+            msr_log(msr, 4, "inject_hashed_response_body: Conv is null.");
             return -1;
         }
 
@@ -1123,6 +1178,7 @@ int inject_hashed_response_body(modsec_rec *msr, int elts) {
         msr->stream_output_data = (char *)malloc(msr->stream_output_length+1);
 
         if (msr->stream_output_data == NULL) {
+            msr_log(msr, 4, "inject_hashed_response_body: Stream Output data is NULL.");
             xmlOutputBufferClose(output_buf);
             xmlFreeDoc(msr->crypto_html_tree);
             return -1;
@@ -1132,7 +1188,7 @@ int inject_hashed_response_body(modsec_rec *msr, int elts) {
         memcpy(msr->stream_output_data, xmlOutputBufferGetContent(output_buf), msr->stream_output_length);
 
         if (msr->txcfg->debuglog_level >= 4)
-            msr_log(msr, 4, "inject_hashed_response_body: Copying XML tree from CONV to stream buffer [%d] bytes.", xmlOutputBufferGetSize(output_buf));
+            msr_log(msr, 4, "inject_hashed_response_body: Copying XML tree from CONV to stream buffer [%zu] bytes.", xmlOutputBufferGetSize(output_buf));
 
     }
 
@@ -1143,6 +1199,7 @@ int inject_hashed_response_body(modsec_rec *msr, int elts) {
         if(output_buf->buffer == NULL || output_buf->buffer->use == 0)  {
             xmlOutputBufferClose(output_buf);
             xmlFreeDoc(msr->crypto_html_tree);
+            msr_log(msr, 4, "inject_hashed_response_body: Output buffer is null.");
             msr->of_stream_changed = 0;
             return -1;
         }
@@ -1158,6 +1215,7 @@ int inject_hashed_response_body(modsec_rec *msr, int elts) {
         if (msr->stream_output_data == NULL) {
             xmlOutputBufferClose(output_buf);
             xmlFreeDoc(msr->crypto_html_tree);
+            msr_log(msr, 4, "inject_hashed_response_body: Stream Output is null.");
             return -1;
         }
 
@@ -1174,6 +1232,7 @@ int inject_hashed_response_body(modsec_rec *msr, int elts) {
             xmlOutputBufferClose(output_buf);
             xmlFreeDoc(msr->crypto_html_tree);
             msr->of_stream_changed = 0;
+            msr_log(msr, 4, "inject_hashed_response_body: Stream Output is null.");
             return -1;
         }
 
@@ -1188,6 +1247,7 @@ int inject_hashed_response_body(modsec_rec *msr, int elts) {
         if (msr->stream_output_data == NULL) {
             xmlOutputBufferClose(output_buf);
             xmlFreeDoc(msr->crypto_html_tree);
+            msr_log(msr, 4, "inject_hashed_response_body: Stream Output Data is null.");
             return -1;
         }
 
@@ -1245,8 +1305,11 @@ char *do_hash_link(modsec_rec *msr, char *link, int type)  {
                 }
 
             if(msr->txcfg->crypto_key_add == HASH_KEYONLY)
+#ifdef __NetBSD__
+                hash_value =  mschmac(msr, msr->txcfg->crypto_key, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#else
                 hash_value =  hmac(msr, msr->txcfg->crypto_key, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
-
+#endif
             if(msr->txcfg->crypto_key_add == HASH_SESSIONID)  {
                 if(msr->sessionid == NULL || strlen(msr->sessionid) == 0)   {
 #if AP_SERVER_MAJORVERSION_NUMBER > 1 && AP_SERVER_MINORVERSION_NUMBER > 2
@@ -1257,13 +1320,21 @@ char *do_hash_link(modsec_rec *msr, char *link, int type)  {
                     if (msr->txcfg->debuglog_level >= 4)
                         msr_log(msr, 4, "Session id is empty. Using REMOTE_IP");
                     msr->txcfg->crypto_key_len = strlen(new_pwd);
+#ifdef __NetBSD__
+                    hash_value = mschmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#else
                     hash_value = hmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#endif
                 } else {
                     const char *new_pwd = apr_psprintf(msr->mp,"%s%s", msr->txcfg->crypto_key, msr->sessionid);
                     if (msr->txcfg->debuglog_level >= 4)
                         msr_log(msr, 4, "Using session id [%s]", msr->sessionid);
                     msr->txcfg->crypto_key_len = strlen(new_pwd);
+#ifdef __NetBSD__
+                    hash_value = mschmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#else
                     hash_value = hmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#endif
                 }
             }
 
@@ -1274,7 +1345,11 @@ char *do_hash_link(modsec_rec *msr, char *link, int type)  {
                 const char *new_pwd = apr_psprintf(msr->mp,"%s%s", msr->txcfg->crypto_key, msr->r->connection->remote_ip);
 #endif
                 msr->txcfg->crypto_key_len = strlen(new_pwd);
+#ifdef __NetBSD__
+                hash_value = mschmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#else
                 hash_value = hmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#endif
             }
         } else  {
             return NULL;
@@ -1288,8 +1363,11 @@ char *do_hash_link(modsec_rec *msr, char *link, int type)  {
                 }
 
                 if(msr->txcfg->crypto_key_add == HASH_KEYONLY)
+#ifdef __NetBSD__
+                    hash_value =  mschmac(msr, msr->txcfg->crypto_key, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#else
                     hash_value =  hmac(msr, msr->txcfg->crypto_key, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
-
+#endif
                 if(msr->txcfg->crypto_key_add == HASH_SESSIONID)  {
                     if(msr->sessionid == NULL || strlen(msr->sessionid) == 0)   {
 #if AP_SERVER_MAJORVERSION_NUMBER > 1 && AP_SERVER_MINORVERSION_NUMBER > 2
@@ -1300,13 +1378,21 @@ char *do_hash_link(modsec_rec *msr, char *link, int type)  {
                         if (msr->txcfg->debuglog_level >= 4)
                             msr_log(msr, 4, "Session id is empty. Using REMOTE_IP");
                         msr->txcfg->crypto_key_len = strlen(new_pwd);
+#ifdef __NetBSD__
+                        hash_value = mschmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#else
                         hash_value = hmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#endif
                     } else {
                         const char *new_pwd = apr_psprintf(msr->mp,"%s%s", msr->txcfg->crypto_key, msr->sessionid);
                         if (msr->txcfg->debuglog_level >= 4)
                             msr_log(msr, 4, "Using session id [%s]", msr->sessionid);
                         msr->txcfg->crypto_key_len = strlen(new_pwd);
+#ifdef __NetBSD__
+                        hash_value = mschmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#else
                         hash_value = hmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#endif
                     }
                 }
 
@@ -1317,7 +1403,11 @@ char *do_hash_link(modsec_rec *msr, char *link, int type)  {
                     const char *new_pwd = apr_psprintf(msr->mp,"%s%s", msr->txcfg->crypto_key, msr->r->connection->remote_ip);
 #endif
                     msr->txcfg->crypto_key_len = strlen(new_pwd);
+#ifdef __NetBSD__
+                    hash_value = mschmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#else
                     hash_value = hmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) path_chunk+1, strlen((char*)path_chunk)-1);
+#endif
                 }
             } else  {
                 return NULL;
@@ -1329,8 +1419,11 @@ char *do_hash_link(modsec_rec *msr, char *link, int type)  {
                 }
 
             if(msr->txcfg->crypto_key_add == HASH_KEYONLY)
+#ifdef __NetBSD__
+                hash_value = mschmac(msr, msr->txcfg->crypto_key, msr->txcfg->crypto_key_len, (unsigned char *) link+1, strlen((char*)link)-1);
+#else
                 hash_value = hmac(msr, msr->txcfg->crypto_key, msr->txcfg->crypto_key_len, (unsigned char *) link+1, strlen((char*)link)-1);
-
+#endif
             if(msr->txcfg->crypto_key_add == HASH_SESSIONID)  {
                 if(msr->sessionid == NULL || strlen(msr->sessionid) == 0)   {
 #if AP_SERVER_MAJORVERSION_NUMBER > 1 && AP_SERVER_MINORVERSION_NUMBER > 2
@@ -1341,13 +1434,21 @@ char *do_hash_link(modsec_rec *msr, char *link, int type)  {
                     if (msr->txcfg->debuglog_level >= 4)
                         msr_log(msr, 4, "Session id is empty. Using REMOTE_IP");
                     msr->txcfg->crypto_key_len = strlen(new_pwd);
+#ifdef __NetBSD__
+                    hash_value = mschmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) link+1, strlen((char*)link)-1);
+#else
                     hash_value = hmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) link+1, strlen((char*)link)-1);
+#endif
                 } else  {
                     const char *new_pwd = apr_psprintf(msr->mp,"%s%s", msr->txcfg->crypto_key, msr->sessionid);
                     if (msr->txcfg->debuglog_level >= 4)
                         msr_log(msr, 4, "Using session id [%s]", msr->sessionid);
                     msr->txcfg->crypto_key_len = strlen(new_pwd);
+#ifdef __NetBSD__
+                    hash_value = mschmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) link+1, strlen((char*)link)-1);
+#else
                     hash_value = hmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) link+1, strlen((char*)link)-1);
+#endif
                 }
             }
 
@@ -1358,7 +1459,11 @@ char *do_hash_link(modsec_rec *msr, char *link, int type)  {
                 const char *new_pwd = apr_psprintf(msr->mp,"%s%s", msr->txcfg->crypto_key, msr->r->connection->remote_ip);
 #endif
                 msr->txcfg->crypto_key_len = strlen(new_pwd);
+#ifdef __NetBSD__
+                hash_value = mschmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) link+1, strlen((char*)link)-1);
+#else
                 hash_value = hmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) link+1, strlen((char*)link)-1);
+#endif
             }
 
         }
@@ -1379,11 +1484,15 @@ char *do_hash_link(modsec_rec *msr, char *link, int type)  {
             relative_link = relative_uri+1;
 
             if (msr->txcfg->debuglog_level >= 4)    {
-                msr_log(msr, 4, "Signing data [%s] size %d", relative_link, strlen(relative_link));
+                msr_log(msr, 4, "Signing data [%s] size %zu", relative_link, strlen(relative_link));
                 }
 
             if(msr->txcfg->crypto_key_add == HASH_KEYONLY)
+#ifdef __NetBSD__
+                hash_value = mschmac(msr, msr->txcfg->crypto_key, msr->txcfg->crypto_key_len, (unsigned char *) relative_link, strlen((char*)relative_link));
+#else
                 hash_value = hmac(msr, msr->txcfg->crypto_key, msr->txcfg->crypto_key_len, (unsigned char *) relative_link, strlen((char*)relative_link));
+#endif
 
             if(msr->txcfg->crypto_key_add == HASH_SESSIONID)  {
                 if(msr->sessionid == NULL || strlen(msr->sessionid) == 0)   {
@@ -1395,13 +1504,21 @@ char *do_hash_link(modsec_rec *msr, char *link, int type)  {
                     if (msr->txcfg->debuglog_level >= 4)
                         msr_log(msr, 4, "Session id is empty. Using REMOTE_IP");
                     msr->txcfg->crypto_key_len = strlen(new_pwd);
+#ifdef __NetBSD__
+                    hash_value = mschmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) relative_link, strlen((char*)relative_link));
+#else
                     hash_value = hmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) relative_link, strlen((char*)relative_link));
+#endif
                 } else {
                     const char *new_pwd = apr_psprintf(msr->mp,"%s%s", msr->txcfg->crypto_key, msr->sessionid);
                     if (msr->txcfg->debuglog_level >= 4)
                         msr_log(msr, 4, "Using session id [%s]", msr->sessionid);
                     msr->txcfg->crypto_key_len = strlen(new_pwd);
+#ifdef __NetBSD__
+                    hash_value = mschmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) relative_link, strlen((char*)relative_link));
+#else
                     hash_value = hmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) relative_link, strlen((char*)relative_link));
+#endif
                 }
             }
 
@@ -1412,7 +1529,11 @@ char *do_hash_link(modsec_rec *msr, char *link, int type)  {
                 const char *new_pwd = apr_psprintf(msr->mp,"%s%s", msr->txcfg->crypto_key, msr->r->connection->remote_ip);
 #endif
                 msr->txcfg->crypto_key_len = strlen(new_pwd);
+#ifdef __NetBSD__
+                hash_value = mschmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) relative_link, strlen((char*)relative_link));
+#else
                 hash_value = hmac(msr, new_pwd, msr->txcfg->crypto_key_len, (unsigned char *) relative_link, strlen((char*)relative_link));
+#endif
             }
 
         link = relative_uri;
