@@ -1,6 +1,6 @@
 /*
 * ModSecurity for Apache 2.x, http://www.modsecurity.org/
-* Copyright (c) 2004-2011 Trustwave Holdings, Inc. (http://www.trustwave.com/)
+* Copyright (c) 2004-2013 Trustwave Holdings, Inc. (http://www.trustwave.com/)
 *
 * You may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
@@ -79,7 +79,11 @@ static int lua_restore(lua_State *L, msc_script *script) {
     dumpr.script = script;
     dumpr.index = 0;
 
+#if LUA_VERSION_NUM > 501
+    return lua_load(L, dump_reader, &dumpr, script->name, NULL);
+#else
     return lua_load(L, dump_reader, &dumpr, script->name);
+#endif
 }
 
 /**
@@ -90,7 +94,11 @@ char *lua_compile(msc_script **script, const char *filename, apr_pool_t *pool) {
     msc_lua_dumpw_t dump;
 
     /* Initialise state. */
+#if LUA_VERSION_NUM > 501
+    L = luaL_newstate();
+#else
     L = lua_open();
+#endif
     luaL_openlibs(L);
 
     /* Find script. */
@@ -103,8 +111,11 @@ char *lua_compile(msc_script **script, const char *filename, apr_pool_t *pool) {
     dump.pool = pool;
     dump.parts = apr_array_make(pool, 128, sizeof(msc_script_part *));
 
+#if LUA_VERSION_NUM >= 503
+    lua_dump(L, dump_writer, &dump, 0);
+#else
     lua_dump(L, dump_writer, &dump);
-
+#endif
     (*script) = apr_pcalloc(pool, sizeof(msc_script));
     (*script)->name = filename;
     (*script)->parts = dump.parts;
@@ -154,7 +165,11 @@ static apr_array_header_t *resolve_tfns(lua_State *L, int idx, modsec_rec *msr,
     if (lua_isuserdata(L, idx) || lua_isnoneornil(L, idx)) { /* No second parameter */
         return tfn_arr;
     } else if (lua_istable(L, idx)) { /* Is the second parameter an array? */
+#if LUA_VERSION_NUM > 501
+        int i, n = lua_rawlen(L, idx);
+#else
         int i, n = lua_objlen(L, idx);
+#endif
 
         for(i = 1; i <= n; i++) {
             lua_rawgeti(L, idx, i);
@@ -405,19 +420,32 @@ int lua_execute(msc_script *script, char *param, modsec_rec *msr, msre_rule *rul
     time_before = apr_time_now();
 
 #ifdef CACHE_LUA
+
     L = msr->L;
     rc = lua_gettop(L);
     if(rc)
         lua_pop(L, rc);
+
 #else
+
     /* Create new state. */
+#if LUA_VERSION_NUM == 502 || LUA_VERSION_NUM == 503 || LUA_VERSION_NUM == 501
+    L = luaL_newstate();
+#elif LUA_VERSION_NUM == 500
     L = lua_open();
+#else
+#error We are only tested under Lua 5.0, 5.1, 5.2, or 5.3.
+#endif
     luaL_openlibs(L);
+
 #endif
 
     if(L == NULL)
         return -1;
 
+    luaL_newmetatable(L, "luaL_msc");
+    lua_newtable(L);
+
     /* Associate msr with the state. */
     lua_pushlightuserdata(L, (void *)msr);
     lua_setglobal(L, "__msr");
@@ -429,7 +457,15 @@ int lua_execute(msc_script *script, char *param, modsec_rec *msr, msre_rule *rul
     }
 
     /* Register functions. */
+#if LUA_VERSION_NUM == 500 || LUA_VERSION_NUM == 501
     luaL_register(L, "m", mylib);
+#elif LUA_VERSION_NUM == 502 || LUA_VERSION_NUM == 503
+    luaL_setfuncs(L, mylib, 0);
+#else
+#error We are only tested under Lua 5.0, 5.1, 5.2, or 5.3.
+#endif
+
+    lua_setglobal(L, "m");
 
     rc = lua_restore(L, script);
     if (rc) {
