@@ -1,27 +1,30 @@
 /*
- * ModSecurity for Apache 2.x, http://www.modsecurity.org/
- * Copyright (c) 2004-2010 Trustwave Holdings, Inc. (http://www.trustwave.com/)
- *
- * This product is released under the terms of the General Public Licence,
- * version 2 (GPLv2). Please refer to the file LICENSE (included with this
- * distribution) which contains the complete text of the licence.
- *
- * There are special exceptions to the terms and conditions of the GPL
- * as it is applied to this software. View the full text of the exception in
- * file MODSECURITY_LICENSING_EXCEPTION in the directory of this software
- * distribution.
- *
- * If any of the files related to licensing are missing or if you have any
- * other questions related to licensing please contact Trustwave Holdings, Inc.
- * directly using the email address support@trustwave.com.
- *
- */
+* ModSecurity for Apache 2.x, http://www.modsecurity.org/
+* Copyright (c) 2004-2022 Trustwave Holdings, Inc. (http://www.trustwave.com/)
+*
+* You may not use this file except in compliance with
+* the License.  You may obtain a copy of the License at
+*
+*     http://www.apache.org/licenses/LICENSE-2.0
+*
+* If any of the files related to licensing are missing or if you have any
+* other questions related to licensing please contact Trustwave Holdings, Inc.
+* directly using the email address security@modsecurity.org.
+*/
+
 #ifndef _MSC_PCRE_H_
 #define _MSC_PCRE_H_
 
 typedef struct msc_regex_t msc_regex_t;
 
+#ifdef WITH_PCRE2
+#define PCRE2_CODE_UNIT_WIDTH 8
+#include "pcre2.h"
+#else
 #include "pcre.h"
+#endif
+
+#ifndef WITH_PCRE2
 
 #ifndef PCRE_ERROR_MATCHLIMIT
 /* Define for compile, but not valid in this version of PCRE. */
@@ -33,17 +36,26 @@ typedef struct msc_regex_t msc_regex_t;
 #define PCRE_ERROR_RECURSIONLIMIT (-21)
 #endif /* PCRE_ERROR_RECURSIONLIMIT */
 
+#endif
+
 #include "apr_general.h"
 #include "modsecurity.h"
 
 struct msc_regex_t {
+#ifdef WITH_PCRE2
+    pcre2_code          *re;
+    pcre2_match_context *match_context;
+#ifdef WITH_PCRE_JIT
+    int                 jit_compile_rc;
+#endif
+
+#else
     void            *re;
     void            *pe;
+#endif
     const char      *pattern;
 };
 
-apr_status_t DSOLOCAL msc_pcre_cleanup(msc_regex_t *regex);
-
 void DSOLOCAL *msc_pregcomp_ex(apr_pool_t *pool, const char *pattern, int options,
                                const char **_errptr, int *_erroffset,
                                int match_limit, int match_limit_recursion);
