@@ -127,6 +127,21 @@ apr_status_t modsecurity_request_body_start(modsec_rec *msr, char **error_msg) {
                 msr_log(msr, 2, "%s", *error_msg);
             }
         }
+        else if (strcmp(msr->msc_reqbody_processor, "JSON") == 0) {
+#ifdef WITH_YAJL
+            if (json_init(msr, &my_error_msg) < 0) {
+                *error_msg = apr_psprintf(msr->mp, "JSON parsing error (init): %s", my_error_msg);
+                msr->msc_reqbody_error = 1;
+                msr->msc_reqbody_error_msg = my_error_msg;
+                msr_log(msr, 2, "%s", *error_msg);
+            }
+#else
+            *error_msg = apr_psprintf(msr->mp, "JSON support was not enabled");
+            msr->msc_reqbody_error = 1;
+            msr->msc_reqbody_error_msg = my_error_msg;
+            msr_log(msr, 2, "%s", *error_msg);
+#endif
+        }
         else if (strcmp(msr->msc_reqbody_processor, "URLENCODED") == 0) {
             /* Do nothing, URLENCODED processor does not support streaming yet. */
         }
@@ -344,6 +359,25 @@ apr_status_t modsecurity_request_body_store(modsec_rec *msr,
                 msr_log(msr, 2, "%s", *error_msg);
             }
         }
+        else if (strcmp(msr->msc_reqbody_processor, "JSON") == 0) {
+            /* Increase per-request data length counter. */
+            msr->msc_reqbody_no_files_length += length;
+
+            /* Process data as JSON. */
+#ifdef WITH_YAJL
+            if (json_process_chunk(msr, data, length, &my_error_msg) < 0) {
+                *error_msg = apr_psprintf(msr->mp, "JSON parsing error: %s", my_error_msg);
+                msr->msc_reqbody_error = 1;
+                msr->msc_reqbody_error_msg = *error_msg;
+                msr_log(msr, 2, "%s", *error_msg);
+            }
+#else
+            *error_msg = apr_psprintf(msr->mp, "JSON support was not enabled");
+            msr->msc_reqbody_error = 1;
+            msr->msc_reqbody_error_msg = *error_msg;
+            msr_log(msr, 2, "%s", *error_msg);
+#endif
+        }
         else if (strcmp(msr->msc_reqbody_processor, "URLENCODED") == 0) {
             /* Increase per-request data length counter. */
             msr->msc_reqbody_no_files_length += length;
@@ -362,13 +396,14 @@ apr_status_t modsecurity_request_body_store(modsec_rec *msr,
 
     /* Check that we are not over the request body no files limit. */
     if (msr->msc_reqbody_no_files_length >= (unsigned long) msr->txcfg->reqbody_no_files_limit) {
-
         *error_msg = apr_psprintf(msr->mp, "Request body no files data length is larger than the "
                 "configured limit (%ld).", msr->txcfg->reqbody_no_files_limit);
         if (msr->txcfg->debuglog_level >= 1) {
             msr_log(msr, 1, "%s", *error_msg);
         }
 
+        msr->msc_reqbody_error = 1;
+
         if ((msr->txcfg->is_enabled == MODSEC_ENABLED) && (msr->txcfg->if_limit_action == REQUEST_BODY_LIMIT_ACTION_REJECT))   {
             return -5;
         } else if (msr->txcfg->if_limit_action == REQUEST_BODY_LIMIT_ACTION_PARTIAL)  {
@@ -377,7 +412,6 @@ apr_status_t modsecurity_request_body_store(modsec_rec *msr,
         }
     }
 
-
     /* Store data. */
     if (msr->msc_reqbody_storage == MSC_REQBODY_MEMORY) {
         return modsecurity_request_body_store_memory(msr, data, length, error_msg);
@@ -394,10 +428,16 @@ apr_status_t modsecurity_request_body_store(modsec_rec *msr,
 }
 
 apr_status_t modsecurity_request_body_to_stream(modsec_rec *msr, const char *buffer, int buflen, char **error_msg) {
+#ifndef MSC_LARGE_STREAM_INPUT
     char *stream_input_body = NULL;
     char *data = NULL;
     int first_pkt = 0;
+#else
+    apr_size_t allocate_length = 0;
+    char* allocated = NULL;
+#endif
 
+#ifndef MSC_LARGE_STREAM_INPUT
     if(msr->stream_input_data == NULL)  {
         msr->stream_input_data = (char *)calloc(sizeof(char), msr->stream_input_length + 1);
         first_pkt = 1;
@@ -440,9 +480,63 @@ apr_status_t modsecurity_request_body_to_stream(modsec_rec *msr, const char *buf
         free(data);
         data = NULL;
     }
+#else
+    if (msr->stream_input_data == NULL)  {
+        // Is the request body length known beforehand? (requests that are not Transfer-Encoding: chunked)
+        if (msr->request_content_length > 0) {
+            // Use min of Content-Length and SecRequestBodyLimit
+            allocate_length = msr->request_content_length < msr->txcfg->reqbody_limit ? msr->request_content_length : msr->txcfg->reqbody_limit;
+        }
+        else {
+            // We don't know how this request is going to be, so hope for just buflen to begin with (requests that are Transfer-Encoding: chunked)
+            allocate_length = buflen;
+        }
+
+        allocated = (char*) calloc(allocate_length, sizeof(char));
+        if (allocated) {
+            msr->stream_input_data = allocated;
+            msr->stream_input_allocated_length = allocate_length;
+        }
+        else {
+            *error_msg = apr_psprintf(
+                msr->mp,
+                "Unable to allocate memory to hold request body on stream. Asked for %" APR_SIZE_T_FMT " bytes.",
+                allocate_length);
+            return -1;
+        }
+    }
+    else {
+        // Do we need to expand the space we have previously allocated?
+        if ((msr->stream_input_length + buflen) > msr->stream_input_allocated_length) {
+            // If this becomes a hotspot again, consider increasing by some percent extra each time, for fewer reallocs
+            allocate_length = msr->stream_input_length + buflen;
+
+            allocated = (char*) realloc(msr->stream_input_data, allocate_length);
+            if (allocated) {
+                msr->stream_input_data = allocated;
+                msr->stream_input_allocated_length = allocate_length;
+            }
+            else {
+                *error_msg = apr_psprintf(
+                    msr->mp,
+                    "Unable to reallocate memory to hold request body on stream. Asked for %" APR_SIZE_T_FMT " bytes.",
+                    allocate_length);
+                free(msr->stream_input_data);
+                msr->stream_input_data = NULL;
+                msr->stream_input_length = 0;
+                msr->stream_input_allocated_length = 0;
+                return -1;
+            }
+        }
+    }
+    // Append buffer to msr->stream_input_data
+    memcpy(msr->stream_input_data + msr->stream_input_length, buffer, buflen);
+    msr->stream_input_length += buflen;
+#endif
 
     return 1;
 }
+
 /**
  * Replace a bunch of chunks holding a request body with a single large chunk.
  */
@@ -562,6 +656,19 @@ apr_status_t modsecurity_request_body_end(modsec_rec *msr, char **error_msg) {
     /* Note that we've read the body. */
     msr->msc_reqbody_read = 1;
 
+
+    /* Check that we are not over the request body no files limit. */
+    if (msr->msc_reqbody_no_files_length >= (unsigned long)msr->txcfg->reqbody_no_files_limit) {
+        *error_msg = apr_psprintf(msr->mp, "Request body no files data length is larger than the "
+            "configured limit (%ld).", msr->txcfg->reqbody_no_files_limit);
+        if (msr->txcfg->debuglog_level >= 1) {
+            msr_log(msr, 1, "%s", *error_msg);
+        }
+
+        return -5;
+    }
+
+
     /* Finalise body processing. */
     if ((msr->msc_reqbody_processor != NULL) && (msr->msc_reqbody_error == 0)) {
         char *my_error_msg = NULL;
@@ -601,6 +708,24 @@ apr_status_t modsecurity_request_body_end(modsec_rec *msr, char **error_msg) {
                 return -1;
             }
         }
+        else if (strcmp(msr->msc_reqbody_processor, "JSON") == 0) {
+#ifdef WITH_YAJL
+            if (json_complete(msr, &my_error_msg) < 0 && msr->msc_reqbody_length > 0) {
+                *error_msg = apr_psprintf(msr->mp, "JSON parser error: %s", my_error_msg);
+                msr->msc_reqbody_error = 1;
+                msr->msc_reqbody_error_msg = *error_msg;
+                msr_log(msr, 2, "%s", *error_msg);
+                 return -1;
+             }
+#else
+            *error_msg = apr_psprintf(msr->mp, "JSON support was not enabled");
+            msr->msc_reqbody_error = 1;
+            msr->msc_reqbody_error_msg = *error_msg;
+            msr_log(msr, 2, "%s", *error_msg);
+            return -1;
+#endif
+
+        }
         else if (strcmp(msr->msc_reqbody_processor, "URLENCODED") == 0) {
             return modsecurity_request_body_end_urlencoded(msr, error_msg);
         }
@@ -836,6 +961,11 @@ apr_status_t modsecurity_request_body_clear(modsec_rec *msr, char **error_msg) {
                 const char *put_filename = NULL;
                 const char *put_basename = NULL;
 
+                if (strcmp(msr->txcfg->upload_dir, msr->txcfg->tmp_dir) == 0) {
+                    msr_log(msr, 4, "Not moving file to identical location.");
+                    goto nullify;
+                }
+
                 /* Construct the new filename. */
                 put_basename = file_basename(msr->msc_reqbody_mp, msr->msc_reqbody_filename);
                 if (put_basename == NULL) {
@@ -881,6 +1011,8 @@ apr_status_t modsecurity_request_body_clear(modsec_rec *msr, char **error_msg) {
                     msr->msc_reqbody_filename);
             }
 
+nullify:
+
             msr->msc_reqbody_filename = NULL;
         }
     }
