@@ -1,6 +1,6 @@
 /*
 * ModSecurity for Apache 2.x, http://www.modsecurity.org/
-* Copyright (c) 2004-2011 Trustwave Holdings, Inc. (http://www.trustwave.com/)
+* Copyright (c) 2004-2022 Trustwave Holdings, Inc. (http://www.trustwave.com/)
 *
 * You may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
@@ -14,17 +14,28 @@
 
 #include "msc_xml.h"
 
+static xmlParserInputBufferPtr
+xml_unload_external_entity(const char *URI, xmlCharEncoding enc)    {
+    return NULL;
+}
+
 
 /**
  * Initialise XML parser.
  */
 int xml_init(modsec_rec *msr, char **error_msg) {
+    xmlParserInputBufferCreateFilenameFunc entity;
+
     if (error_msg == NULL) return -1;
     *error_msg = NULL;
 
     msr->xml = apr_pcalloc(msr->mp, sizeof(xml_data));
     if (msr->xml == NULL) return -1;
 
+    if(msr->txcfg->xml_external_entity == 0)    {
+        entity = xmlParserInputBufferCreateFilenameDefault(xml_unload_external_entity);
+    }
+
     return 1;
 }
 
@@ -126,6 +137,16 @@ int xml_complete(modsec_rec *msr, char **error_msg) {
  * Frees the resources used for XML parsing.
  */
 apr_status_t xml_cleanup(modsec_rec *msr) {
+    if (msr->xml->parsing_ctx != NULL) {
+        if (msr->xml->parsing_ctx->myDoc) {
+            xmlFreeDoc(msr->xml->parsing_ctx->myDoc);
+            if (msr->xml->parsing_ctx->myDoc == msr->xml->doc) {
+                msr->xml->doc = NULL;
+            }
+        }
+        xmlFreeParserCtxt(msr->xml->parsing_ctx);
+        msr->xml->parsing_ctx = NULL;
+    }
     if (msr->xml->doc != NULL) {
         xmlFreeDoc(msr->xml->doc);
         msr->xml->doc = NULL;
