@@ -1,6 +1,6 @@
 /*
 * ModSecurity for Apache 2.x, http://www.modsecurity.org/
-* Copyright (c) 2004-2011 Trustwave Holdings, Inc. (http://www.trustwave.com/)
+* Copyright (c) 2004-2022 Trustwave Holdings, Inc. (http://www.trustwave.com/)
 *
 * You may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
@@ -28,7 +28,12 @@
 #if APR_HAVE_UNISTD_H
 #include <unistd.h>         /* for getpid() */
 #endif
+#ifdef WITH_PCRE2
+#define PCRE2_CODE_UNIT_WIDTH 8
+#include <pcre2.h>
+#else
 #include <pcre.h>
+#endif
 #include <curl/curl.h>
 #include <fcntl.h>
 #include <sys/stat.h>
@@ -96,7 +101,7 @@ do { \
 static const char logline_pattern[] =
     "^(\\S+)"
     "\\ (\\S+)\\ (\\S+)\\ (\\S+)"
-    "\\ \\[([^:]+):(\\d+:\\d+:\\d+)\\ ([^\\]]+)\\]"
+    "\\ \\[([^:]+):(\\d+:\\d+:\\d+(?:[.]\\d+)?)\\ ([^\\]]+)\\]"
     "\\ \"(.*)\""
     "\\ (\\d+)\\ (\\S+)"
     "\\ \"(.*)\"\\ \"(.*)\""
@@ -147,7 +152,13 @@ static int                    keep_alive = 150;           /* Not used yet. */
 static int                    keep_alive_timeout = 300;   /* Not used yet. */
 static int                    keep_entries = 0;
 static const char            *log_repository = NULL;
+#ifdef WITH_PCRE2
+static pcre2_code            *logline_regex = NULL;
+static pcre2_code            *requestline_regex = NULL;
+#else
 static void                  *logline_regex = NULL;
+static void                  *requestline_regex = NULL;
+#endif
 static int                    max_connections = 10;
 static int                    max_worker_requests = 1000;
 static apr_global_mutex_t    *gmutex = NULL;
@@ -157,8 +168,10 @@ static apr_pool_t            *thread_pool = NULL;
 static apr_pool_t            *recv_pool = NULL;
 static apr_array_header_t    *queue = NULL;
 static const char            *queue_path = NULL;
+static int                    ssl_validation = 0;
+static int                    tlsprotocol = 1;
+static curl_version_info_data* curlversion = NULL;
 /* static apr_time_t             queue_time = 0; */
-static void                  *requestline_regex = NULL;
 static int                    running = 0;
 static const char            *sensor_password = NULL;
 static const char            *sensor_username = NULL;
@@ -809,6 +822,26 @@ static void init_configuration(void)
         startup_delay = atoi(s);
     }
 
+    /* TLS Protocol - TLSv1(0) TLSv1.1(1) TLSv1.2(2) (SSLv3 not supported) */
+    s = apr_table_get(conf, "TLSProtocol");
+    if (s != NULL) {
+    	int num = atoi(s);
+    	switch (num) {
+    	case 0:
+    		tlsprotocol = 0;
+    		break;
+    	case 1:
+    		tlsprotocol = 1;
+    		break;
+    	case 2:
+    		tlsprotocol = 2;
+    		break;
+    	default:
+    		tlsprotocol = 2; /* Default is TLSv1.2 */
+    	}
+    }
+    curlversion = curl_version_info(CURLVERSION_NOW);
+
     if ( startup_delay > 0 ) {
         error_log(LOG_NOTICE, NULL,
                   "Delaying execution for %dms.", startup_delay);
@@ -823,6 +856,8 @@ static void init_configuration(void)
     error_log(LOG_DEBUG2, NULL, "ErrorLog=%s", error_log_path);
     error_log(LOG_DEBUG2, NULL, "ErrorLogLevel=%d", error_log_level);
     error_log(LOG_DEBUG2, NULL, "StartupDelay=%d", startup_delay);
+    error_log(LOG_DEBUG2, NULL, "TLSProtocol=%d", tlsprotocol);
+    error_log(LOG_DEBUG2, NULL, "cURL version=%s",  curlversion->version);
 
     s = apr_table_get(conf, "CheckpointInterval");
     if (s != NULL) {
@@ -831,6 +866,20 @@ static void init_configuration(void)
                   "CheckpointInterval=%d", checkpoint_interval);
     }
 
+    s = apr_table_get(conf, "InsecureNoCheckCert");
+    if (s != NULL) {
+        int num = atoi(s);
+        if (num)
+        {
+            ssl_validation = 0;
+        }
+        else
+        {
+            ssl_validation = 1;
+        }
+        error_log(LOG_DEBUG2, NULL, "InsecureNoCheckCert=%d", num);
+    }
+
     s = apr_table_get(conf, "QueuePath");
     if (s != NULL) {
         queue_path = file_path(s);
@@ -1167,6 +1216,12 @@ static void logc_init(void)
     apr_status_t rc = 0;
     const char *errptr = NULL;
     int i, erroffset;
+    /* cURL major, minor and patch version */
+    short cmaj, cmin, cpat = 0;
+#ifdef WITH_PCRE2
+    int pcre2_errorcode = 0;
+    PCRE2_SIZE pcre2_erroffset = 0;
+#endif
 
     queue = apr_array_make(pool, 64, sizeof(entry_t *));
     if (queue == NULL) {
@@ -1216,10 +1271,47 @@ static void logc_init(void)
         curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, (char *)NULL);
         curl_easy_setopt(curl, CURLOPT_URL, console_uri);
         curl_easy_setopt(curl, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
-        curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, FALSE);
-        curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, 0);
-        /* SSLv3 works better overall as some servers have issues with TLS */
-        curl_easy_setopt(curl, CURLOPT_SSLVERSION, CURL_SSLVERSION_SSLv3);
+
+        if (ssl_validation)
+        {
+            curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, 1);
+            curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, 1);
+        }
+        else
+        {
+            curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, FALSE);
+            curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, 0);
+        }
+
+
+        /* Seems like CURL_SSLVERSION_TLSv1_2 is not supported on libcurl
+         * < v7.34.0
+         *
+         * version_num is a 24 bit number created like this:
+         * <8 bits major number> | <8 bits minor number> | <8 bits patch number>.
+         */
+        switch (tlsprotocol) {
+        case 0:
+        	curl_easy_setopt(curl, CURLOPT_SSLVERSION, CURL_SSLVERSION_TLSv1_0);
+        	break;
+        case 1:
+        	curl_easy_setopt(curl, CURLOPT_SSLVERSION, CURL_SSLVERSION_TLSv1_1);
+        	break;
+        case 2:
+        	curl_easy_setopt(curl, CURLOPT_SSLVERSION, CURL_SSLVERSION_TLSv1_2);
+        	break;
+        default:
+        	curl_easy_setopt(curl, CURLOPT_SSLVERSION, CURL_SSLVERSION_TLSv1_2);
+        	break;
+        }
+        cmaj = curlversion->version_num >> 16;
+        cmin = (curlversion->version_num & 0x00ff00) >> 8;
+        cpat = (curlversion->version_num & 0x0000ff);
+        /* If cURL version < v7.34.0, use TLS v1.x */
+        if (cmaj <= 7 && cmin < 34) {
+        	curl_easy_setopt(curl, CURLOPT_SSLVERSION, CURL_SSLVERSION_TLSv1);
+        }
+
         curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, 15);
         curl_easy_setopt(curl, CURLOPT_NOSIGNAL, TRUE);
         curl_easy_setopt(curl, CURLOPT_HEADER, TRUE);
@@ -1229,16 +1321,30 @@ static void logc_init(void)
         *(CURL **)apr_array_push(curl_handles) = curl;
     }
 
+    if (cmaj <= 7 && cmin < 34) {
+    	error_log(LOG_DEBUG2, NULL, "TLSv1.2 is unsupported in cURL %d.%d.%d",  cmaj, cmin, cpat);
+    }
+
+#ifdef WITH_PCRE2
+    logline_regex = pcre2_compile(logline_pattern, PCRE2_ZERO_TERMINATED, PCRE2_CASELESS,
+                                 &pcre2_errorcode, &pcre2_erroffset, NULL);
+#else
     logline_regex = pcre_compile(logline_pattern, PCRE_CASELESS,
                                  &errptr, &erroffset, NULL);
+#endif
     if (logline_regex == NULL) {
         error_log(LOG_ERROR, NULL,
                   "Failed to compile pattern: %s\n", logline_pattern);
         logc_shutdown(1);
     }
 
-    requestline_regex = pcre_compile(requestline_pattern,
-                                     PCRE_CASELESS, &errptr, &erroffset, NULL);
+#ifdef WITH_PCRE2
+    requestline_regex = pcre2_compile(requestline_pattern, PCRE2_ZERO_TERMINATED, PCRE2_CASELESS,
+                                     &pcre2_errorcode, &pcre2_erroffset, NULL);
+#else
+    requestline_regex = pcre_compile(requestline_pattern, PCRE_CASELESS,
+                                     &errptr, &erroffset, NULL);
+#endif
     if (requestline_regex == NULL) {
         error_log(LOG_ERROR, NULL,
                   "Failed to compile pattern: %s\n", requestline_pattern);
@@ -1349,6 +1455,9 @@ static void * APR_THREAD_FUNC thread_worker(apr_thread_t *thread, void *data)
     apr_status_t rc;
     apr_finfo_t finfo;
     int capturevector[CAPTUREVECTORSIZE];
+#ifdef WITH_PCRE2
+    pcre2_match_data *pcre2_match_data = NULL;
+#endif
     int take_new = 1;
     apr_pool_t *tpool;
     struct curl_slist *headerlist = NULL;
@@ -1454,9 +1563,24 @@ static void * APR_THREAD_FUNC thread_worker(apr_thread_t *thread, void *data)
             num_requests++;
         }
 
+#ifdef WITH_PCRE2
+        pcre2_match_data  = pcre2_match_data_create_from_pattern(logline_regex, NULL);
+        rc = pcre2_match(logline_regex, entry->line, entry->line_size, 0, 0,
+            pcre2_match_data, NULL);
+	if (rc > 0) {
+            PCRE2_SIZE *pcre2_ovector = pcre2_get_ovector_pointer(pcre2_match_data);
+            for (int i = 0; i < rc; i++) {
+                capturevector[2*i] = pcre2_ovector[2*i];
+                capturevector[2*i+1] = pcre2_ovector[2*i+1];
+            }
+        }
+        pcre2_match_data_free(pcre2_match_data);
+        if (rc == PCRE2_ERROR_NOMATCH) {
+#else
         rc = pcre_exec(logline_regex, NULL, entry->line, entry->line_size, 0, 0,
             capturevector, CAPTUREVECTORSIZE);
-        if (rc == PCRE_ERROR_NOMATCH) { /* No match. */
+        if (rc == PCRE_ERROR_NOMATCH) {
+#endif
             error_log(LOG_WARNING, thread,
                       "Invalid entry (failed to match regex): %s",
                       _log_escape(tpool, entry->line, entry->line_size));
@@ -2210,6 +2334,11 @@ static void usage(void) {
  * Version text.
  */
 static void version(void) {
+#ifdef WITH_PCRE2
+    char pcre2_loaded_version_buffer[80] ={0};
+    char *pcre_loaded_version = pcre2_loaded_version_buffer;
+    pcre2_config(PCRE2_CONFIG_VERSION, pcre_loaded_version);
+#endif
     fprintf(stderr,
             "ModSecurity Log Collector (mlogc) v%s\n", VERSION);
     fprintf(stderr,
@@ -2217,7 +2346,11 @@ static void version(void) {
             "loaded=\"%s\"\n", APR_VERSION_STRING, apr_version_string());
     fprintf(stderr,
             "  PCRE: compiled=\"%d.%d\"; "
+#ifdef WITH_PCRE2
+            "loaded=\"%s\"\n", PCRE2_MAJOR, PCRE2_MINOR, pcre_loaded_version);
+#else
             "loaded=\"%s\"\n", PCRE_MAJOR, PCRE_MINOR, pcre_version());
+#endif
     fprintf(stderr,
             "  cURL: compiled=\"%s\"; "
             "loaded=\"%s\"\n", LIBCURL_VERSION, curl_version());
